#include "DataRecorder.h"

#pragma pack(push, 1)

struct PosInfoDisk
{
    uint32_t head;
    uint32_t tail;
};

struct Record
{
    uint8_t len;
    uint8_t dummy[RECORD_DUMMY_SIZE];
    uint8_t data[RECORD_DATA_SIZE];
    uint16_t crc;
};

struct RecordBlock
{
    Record record[RECORD_PER_BLOCK];
};

#pragma pack(pop)

static RecordBlock headBlock;
static RecordBlock tailBlock;
static RecordBlock tempReadBlock;

int DataRecorder::Init(VFileSystem* fs)
{
    this->fs = fs;
    int ret;
    if (ret = ReadPosition()) {
        return ret;
    }
    if (ret = ReadTailRecordBlock(tail / RECORD_PER_BLOCK)) {
        return ret;
    }
    if (HeadTailSameBlock()) {
        SyncHeadTailBlock();
    } else {
        if (ret = ReadHeadRecordBlock(head / RECORD_PER_BLOCK)) {
            return ret;
        }
    }
    tempReadBlockIndex = -1;
    return 0;
}

int DataRecorder::PushData(uint8_t len, const uint8_t* data)
{
    if (len >= RECORD_DATA_SIZE) {
        return -3;
    }
    tailBlock.record[tail % RECORD_PER_BLOCK].len = len;
    memcpy(tailBlock.record[tail % RECORD_PER_BLOCK].data, data, len);
    if (HeadTailSameBlock()) {
        SyncHeadTailBlock();
    }
    return IncTailIndex();
}

int DataRecorder::PopData(int& index, uint8_t& len, uint8_t* data)
{
    if (head == tail) {
        return -4;
    }
    index = head;
    len = headBlock.record[head % RECORD_PER_BLOCK].len;
    memcpy(data, headBlock.record[head % RECORD_PER_BLOCK].data, len);
    return IncHeadIndex();
}

int DataRecorder::ReadData(int& index, uint8_t& len, uint8_t* data)
{
    if (head == tail) {
        return -4;
    }
    index = head;
    len = headBlock.record[head % RECORD_PER_BLOCK].len;
    memcpy(data, headBlock.record[head % RECORD_PER_BLOCK].data, len);
    return 0;
}

int DataRecorder::ReadDataFromIndex(int index, uint8_t& len, uint8_t* data)
{
    int ret;
    if (index / RECORD_PER_BLOCK != tempReadBlockIndex) {
        if (ret = ReadTempRecordBlock(index / RECORD_PER_BLOCK)) {
            return ret;
        }
        tempReadBlockIndex = index / RECORD_PER_BLOCK;
    }
    len = tempReadBlock.record[index % RECORD_PER_BLOCK].len;
    memcpy(data, tempReadBlock.record[index % RECORD_PER_BLOCK].data, len);
    return 0;
}

int DataRecorder::HeadIndex()
{
    return head;
}

int DataRecorder::TailIndex()
{
    return tail;
}

int DataRecorder::RecordCount()
{
    if (tail >= head) {
        return tail - head;
    } else {
        return RECORD_COUNT + tail - head;
    }
}

int DataRecorder::ForceSave()
{
    int ret;
    if (ret = WriteTailRecordBlock(tail / RECORD_PER_BLOCK)) {
        return ret;
    }
    return WritePosition();
}

int DataRecorder::IncHeadIndex()
{
    head++;
    if (head == RECORD_COUNT) {
        head = 0;
    }
    if (HeadTailSameBlock()) {
        SyncHeadTailBlock();
        return 0;
    }
    if (head % RECORD_PER_BLOCK == 0) {
        int ret;
        if (ret = ReadHeadRecordBlock(head / RECORD_PER_BLOCK)) {
            return ret;
        }
        return WritePosition();
    } else if (head % FORCESAVE_READ_COUNT == 0) {
        return ForceSave();
    }
    return 0;
}

int DataRecorder::IncTailIndex()
{
    tail++;
    if (tail == RECORD_COUNT) {
        tail = 0;
    }
    int ret;
    if (tail == head) {
        if (ret = IncHeadIndex()) {
            return ret;
        }
    }
    if (tail % RECORD_PER_BLOCK == 0) {
        if (ret = WriteTailRecordBlock(tail / RECORD_PER_BLOCK - 1)) {
            return ret;
        }
        if (ret = ReadTailRecordBlock(tail / RECORD_PER_BLOCK)) {
            return ret;
        }
        return WritePosition();
    } else if (tail % FORCESAVE_WRITE_COUNT == 0) {
        return ForceSave();
    }
    return 0;
}

bool DataRecorder::HeadTailSameBlock()
{
    return head / RECORD_PER_BLOCK == tail / RECORD_PER_BLOCK;
}

void DataRecorder::SyncHeadTailBlock()
{
    memcpy(&headBlock, &tailBlock, RECORD_BLOCK_SIZE);
}

int DataRecorder::ReadHeadRecordBlock(int blockIndex)
{
    char filename[14];
    sprintf(filename, "/rec-%04d.dat", blockIndex);
    if (fs->ReadFile(filename, (uint8_t*)&headBlock, RECORD_BLOCK_SIZE)) {
        return -1;
    }
    return 0;
}

int DataRecorder::ReadTailRecordBlock(int blockIndex)
{
    char filename[14];
    sprintf(filename, "/rec-%04d.dat", blockIndex);
    if (fs->ReadFile(filename, (uint8_t*)&tailBlock, RECORD_BLOCK_SIZE)) {
        return -1;
    }
    return 0;
}

int DataRecorder::ReadTempRecordBlock(int blockIndex)
{
    char filename[14];
    sprintf(filename, "/rec-%04d.dat", blockIndex);
    if (fs->ReadFile(filename, (uint8_t*)&tempReadBlock, RECORD_BLOCK_SIZE)) {
        return -1;
    }
    return 0;
}

int DataRecorder::WriteTailRecordBlock(int blockIndex)
{
    char filename[14];
    sprintf(filename, "/rec-%04d.dat", blockIndex);
    if (fs->WriteFile(filename, (uint8_t*)&tailBlock, RECORD_BLOCK_SIZE)) {
        return -1;
    }
    return 0;
}

int DataRecorder::ReadPosition()
{
    PosInfoDisk pos;
    int ret;
    if (ret = fs->ReadFile("/position.dat", (uint8_t*)&pos, sizeof(PosInfoDisk))) {
        return ret;
    }
    head = pos.head;
    tail = pos.tail;
    return 0;
}

int DataRecorder::WritePosition()
{
    PosInfoDisk pos;
    pos.head = head;
    pos.tail = tail;
    int ret;
    if (ret = fs->WriteFile("/position.dat", (uint8_t*)&pos, sizeof(PosInfoDisk))) {
        return ret;
    }
    return 0;
}
